#!/usr/local/bin/perl
#
# Usage:
#   <img src="http://www.exsample.com/counter.cgi?file=jindex.html&width=8" />
#
# Copyright (C) 2005 by Hiroshi Yuki.
# All rights reserved.
# http://www.hyuki.com/
#
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# download from
# http://codezine.jp/a/article.aspx?aid=70
#
# non-compression GIF image creation from XBM data format
# Copyright (C) 2007 by George
# All rights reserved.
# http://www.yk.rim.or.jp/~george/
#
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# 2007.08.29 modified by george@yk.rim.or.jp for creation 
#            non-compression GIF image from XBM data.
#
#use strict;

# digit bitmap data (width = 8, height = 10 dots XBM format)
my %image = (
'0' => "\x3c\x66\x66\x66\x66\x66\x66\x66\x66\x3c",
'1' => "\x30\x38\x30\x30\x30\x30\x30\x30\x30\x30",
'2' => "\x3c\x66\x60\x60\x30\x18\x0c\x06\x06\x7e",
'3' => "\x3c\x66\x60\x60\x38\x60\x60\x60\x66\x3c",
'4' => "\x30\x30\x38\x38\x34\x34\x32\x7e\x30\x78",
'5' => "\x7e\x06\x06\x06\x3e\x60\x60\x60\x66\x3c",
'6' => "\x38\x0c\x06\x06\x3e\x66\x66\x66\x66\x3c",
'7' => "\x7e\x66\x60\x60\x30\x30\x18\x18\x0c\x0c",
'8' => "\x3c\x66\x66\x66\x3c\x66\x66\x66\x66\x3c",
'9' => "\x3c\x66\x66\x66\x66\x7c\x60\x60\x30\x1c",
);


&main;
exit;

sub main {
    my $counter_file = &get_param_file();
    my $width = &get_param_width();
    my $counter = &get_access_counter($counter_file);
    my $string = sprintf("%0" . $width . "d", $counter);
    print "Content-type: image/gif\n";
    print "\n";
    binmode(STDOUT);
    print &concat_total_gif($string);
}

sub get_param_file {
    my $query = $ENV{"QUERY_STRING"};
    if ( $query =~ /file=([a-zA-Z0-9_.]+)/ ) {
	return "lock-$1";
    }
    return "lock";
}

sub get_param_width {
    my $query = $ENV{"QUERY_STRING"};
    if ( $query =~ /width=([0-9])/ ) {
	my $width = $1;
	if ( "4" <= $width &&  $width <= "8" ) {
	    return $width;
	}
    }
    return "6";
}

sub get_access_counter {
    my ($fname) = @_;

    if ( ! -e($fname) ) {
	return 404;
    }

    if (open(FILE, "+< $fname")) {
        flock(FILE, 2); # 2=LOCK_EX
        seek(FILE, 0, 0);
        my $count = <FILE> + 1;
        seek(FILE, 0, 0);
        print FILE "$count\n";
        flock(FILE, 8); # 8=LOCK_UN
        close(FILE);
        return $count;
    } else {
        # File open error.
        return 500;
    }
}

# Create a GIF header according to given width, height.
sub make_header {
    my ($width, $height) = @_;
    my $result = "GIF89a";                  # signature
    $result .= &pack_xy($width, $height);   # local screen width, logical screen height
    $result .= "\x80";                      # monochrome Global Color Table = true
    $result .= "\x00";                      # bgindex
    $result .= "\x00";                      # no aspect ratio
    $result .= "\x00\x00\x00";              # black color
    $result .= "\xFF\xFF\xFF";              # white color
    return $result;
}

# Footer (trailer).
sub make_footer {
    return "\x3B";
}

#----------------------------------------------------------

# Create image (all digits = 1 sub image version. )
sub concat_total_gif {
    my ($string) = @_;
    my $result = '';
    $result .= &make_header(8 * length($string), 10);
    $result .= &make_total_image($string);
    $result .= &make_footer;
    return $result;
}

# Create images. (all digits = 1 sub image version. )
sub make_total_image {
    my ($string) = @_;
    my $result = '';
    $result .= &make_control();
    $result .= &make_total_subimage(0, 0, $string);
    return $result;
}

# Create a GIF subimage. (all digits in one sub image version.)
sub make_total_subimage {
    my ($left, $top, $name) = @_;
	my $image_data = '';
    my $result = '';
	$result .= "\x2C";                  # introducer
    $result .= &pack_xy($left, $top);   # image left position, top position
	$result .= &pack_xy(length($name)*8, 10);     # image Width, Height size
	$result .= "\x00";                  # LCTF/IF/SF/R/SLCT
	$image_data = &make_total_xbm_image($name);
	$result .= &pack_2bpp_image_block( $image_data );
    return $result;
}

# Create XBM image (all digits in one image version)
sub make_total_xbm_image {
	my ($string) = @_;
	my $row = 0;
	my $result = '';
	for ( $row = 0 ; $row < 10 ; $row++ ) {
		foreach ( split(//, $string) ) {
			$result = substr( $image{$_}, $row, 1) . $result;
		}
	}
	return $result;
}

# Create a Graphic Control Extension.
sub make_control {
    my $result = "\x21";                # introducer
    $result .= "\xF9";                  # Label
    $result .= "\x04";                  # size
    $result .= "\x00";                  # Disposal Method / User Input Flag / Transparent Color Flag
	$result .= "\x00\x00";              # Delay Time
	$result .= "\x00";                  # Transparent Color Index
    $result .= "\x00";                  # Block Terminator
    return $result;
}

# Return packed 2 bytes according to given x, y
sub pack_xy {
    my ($x, $y) = @_;
    my $result = '';
    $result .= chr($x % 256);
    $result .= chr(($x / 256) % 256);
    $result .= chr($y % 256);
    $result .= chr(($y / 256) % 256);
    return $result;
}

# Return packed 2bpp(4 colors) bitmap image
# (monoclome uses 2bpp mode)
sub pack_2bpp_image_block {
    my ($bitmap) = @_;
    my $result = '';
	my $bitmap01 = unpack("B*", $bitmap );
	my $bitstream = '';
	my $count = 0;
	my $bits = '';
	foreach ( split ( //, $bitmap01 ) ) {
		# read 1 bit (monoclome data )
		$bits = $_;

		$bitstream = '00' . $bits . $bitstream ;     # write in 3bit
		$count++;
		if ( ($count % 2) == 1 ) {
			$bitstream = '100' . $bitstream ;        # clear code
		}
	}
	$bitstream = $bitstream . '101';              # end stream	
	$result = &pack_bitstream_into_string($bitstream, 3);
	$result = &pack_bytestream_into_datablock($result);

	$result = "\x02" . $result;         # set LZW block size 
	return $result;
}

# convert bitstream into 8bit byte list
# pack right to left
sub pack_bitstream_into_string {
	my ($bits, $bpp) = @_;
	my $tmp = '';
	my $bitmap = '';
	my $result = '';
	my $regex = '.' x $bpp;
	while ( length($bits) > 0 ) {
		# get first $bpp bits from bitstream
		$tmp = substr($bits, 0, $bpp);
		while ( length($tmp) < $bpp ) {
			$tmp = "0" . $tmp;
		}
		$bits = substr($bits, $bpp);

		# add from left side
		$bitmap = $tmp . $bitmap;
		if ( length($bitmap) >= 8 ) {
			$result .= pack("B8", substr($bitmap, length($bitmap)-8, 8) );
			$bitmap = substr($bitmap, 0, length($bitmap)-8);
		}
	}
	return $result;
}

sub pack_bytestream_into_datablock {
	my ($bytestream) = @_;
	my $result = '';
	my $byteblock = '';
	while ( length($bytestream) >= 254 ) {
		$byteblock = substr($bytestream, 0, 254);
		$result .= chr(length($byteblock)) . $byteblock;
		$bytestream = substr($bytestream, 254);
	}
	if ( length($bytestream) != 0 ) {
		$result .= chr(length($bytestream)) . $bytestream;
	}
	$result .= "\x00"; # block terminator
	return $result;
}


1;
